# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
Python templating code
'''
from __future__ import print_function, unicode_literals, division, absolute_import

from ..utils.printing import fill

def get_interpreter():
    return 'python3'

def get_path():
    return 'python'

class_to_dxclass = {
    "file": "dxpy.DXFile",
    "record": "dxpy.DXRecord",
    "applet": "dxpy.DXApplet"
    }

def get_output_fmt(output_param):
    output_class = output_param["class"]
    output_fmt = ''
    if output_class.startswith('array'):
        item_class = output_class[6:]
        if item_class in class_to_dxclass:
            output_fmt = "[dxpy.dxlink(item) for item in " + output_param["name"] + "]"
        else:
            output_fmt = output_param["name"]
    elif output_class in class_to_dxclass:
        output_fmt = "dxpy.dxlink(" + output_param["name"] + ")"
    else:
        output_fmt = output_param["name"]
    return output_fmt

def add_init_input_lines(init_inputs, input_param, may_be_missing):
    if input_param["class"] in class_to_dxclass:
        init_str = "{name} = {dxclass}({name})".format(name=input_param["name"],
                                                       dxclass=class_to_dxclass[input_param["class"]])
    elif input_param["class"].startswith("array:") and input_param["class"][6:] in class_to_dxclass:
        init_str = "{name} = [{dxclass}(item) for item in {name}]".format(name=input_param["name"],
                                                                          dxclass=class_to_dxclass[input_param["class"][6:]])
    else:
        init_str = None

    if init_str is not None:
        if may_be_missing:
            init_inputs.append("if {name} is not None:".format(name=input_param['name']))
            indent = '    '
        else:
            indent = ''
        init_inputs.append(indent + init_str)

def get_strings(app_json,
                required_file_input_names, optional_file_input_names,
                required_file_array_input_names, optional_file_array_input_names,
                file_output_names, dummy_output_hash):
    input_sig_str = ''
    init_inputs_str = ''
    dl_files_str = ''
    ul_files_str = ''
    outputs_str = ''
    args_list = []
    kwargs_list = []
    init_inputs = []
    if "inputSpec" in app_json:
        # Iterate through input parameters and add them to the
        # signature and make their initialization lines
        for input_param in app_json["inputSpec"]:
            may_be_missing = input_param.get("optional") and "default" not in input_param
            if may_be_missing:
                # add as kwarg
                kwargs_list.append(input_param['name'])
            else:
                # argument that will always be present
                args_list.append(input_param["name"])

            # And no matter what, add line(s) for initializing it if present
            add_init_input_lines(init_inputs, input_param, may_be_missing)
        input_sig_str = ", ".join(args_list + [name + '=None' for name in kwargs_list])
    else:
        input_sig_str = "**kwargs"

    if init_inputs:
        init_inputs_str = '\n' + fill('The following line(s) initialize your data object inputs on the platform into dxpy.DXDataObject instances that you can start using immediately.', initial_indent='    # ', subsequent_indent='    # ', width=80)
        init_inputs_str += "\n\n    "
        init_inputs_str += "\n    ".join(init_inputs)
        init_inputs_str += "\n"

    if required_file_input_names or optional_file_input_names or \
       required_file_array_input_names or optional_file_array_input_names:
        dl_files_str = '\n' + fill('The following line(s) download your file inputs to the local file system using variable names for the filenames.', initial_indent='    # ', subsequent_indent='    # ', width=80) + '\n\n'
        if required_file_input_names:
            dl_files_str += "\n".join(['''    dxpy.download_dxfile({name}.get_id(), "{name}")
'''.format(name=name) for name in required_file_input_names])
        if optional_file_input_names:
            dl_files_str += "\n".join(['''    if {name} is not None:
        dxpy.download_dxfile({name}.get_id(), "{name}")
'''.format(name=name) for name in optional_file_input_names])
        if required_file_array_input_names:
            dl_files_str += "\n".join(['''    for i, f in enumerate({name}):
        dxpy.download_dxfile(f.get_id(), "{name}-" + str(i))
'''.format(name=name) for name in required_file_array_input_names])
        if optional_file_array_input_names:
            dl_files_str += "\n".join(['''    if {name} is not None:
        for i, f in enumerate({name}):
            dxpy.download_dxfile(f.get_id(), "{name}-" + str(i))
'''.format(name=name) for name in optional_file_array_input_names])

    if file_output_names:
        ul_files_str = "\n" + fill('''The following line(s) use the Python bindings to upload your file outputs after you have created them on the local file system.  It assumes that you have used the output field name for the filename for each output, but you can change that behavior to suit your needs.''', initial_indent="    # ", subsequent_indent="    # ", width=80)
        ul_files_str +='\n\n    '
        ul_files_str += "\n    ".join(['{name} = dxpy.upload_local_file("{name}")'.format(name=name) for name in file_output_names]) + '\n'

    if 'outputSpec' in app_json and app_json['outputSpec']:
        outputs_str = "    " + "\n    ".join(['output["{name}"] = {value}'.format(name=param["name"], value=get_output_fmt(param)) for param in app_json['outputSpec']]) + '\n'

    return input_sig_str, init_inputs_str, dl_files_str, ul_files_str, outputs_str



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
DXJob Handler
+++++++++++++

Jobs are DNAnexus entities that capture an instantiation of a running
app or applet. They can be created from either
:func:`dxpy.bindings.dxapplet.DXApplet.run` or
:func:`dxpy.bindings.dxapp.DXApp.run` if running an applet or app, or
via :func:`new_dxjob` or :func:`DXJob.new` in the case of an existing
job creating a subjob.

"""

from __future__ import print_function, unicode_literals, division, absolute_import

import os, time

import dxpy
from . import DXObject, DXDataObject, DXJobFailureError, verify_string_dxid
from ..exceptions import DXError
from ..system_requirements import SystemRequirementsDict
from ..utils.local_exec_utils import queue_entry_point
from ..compat import basestring


#########
# DXJob #
#########


def new_dxjob(fn_input, fn_name, name=None, tags=None, properties=None, details=None, instance_type=None,
              depends_on=None, cluster_spec=None, fpga_driver=None, system_requirements=None,
              system_requirements_by_executable=None, nvidia_driver=None, **kwargs):
    '''
    :param fn_input: Function input
    :type fn_input: dict
    :param fn_name: Name of the function to be called
    :type fn_name: string
    :param name: Name for the new job (default is "<parent job name>:<fn_name>")
    :type name: string
    :param tags: Tags to associate with the job
    :type tags: list of strings
    :param properties: Properties to associate with the job
    :type properties: dict with string values
    :param details: Details to set for the job
    :type details: dict or list
    :param instance_type: Instance type on which the job will be run, or a dict mapping function names to instance type requests
    :type instance_type: string or dict
    :param depends_on: List of data objects or jobs to wait that need to enter the "closed" or "done" states, respectively, before the new job will be run; each element in the list can either be a dxpy handler or a string ID
    :type depends_on: list
    :param cluster_spec: a dict mapping function names to cluster spec requests
    :type cluster_spec: dict
    :param fpga_driver: a dict mapping function names to fpga driver requests
    :type fpga_driver: dict
    :param system_requirements: System requirement single mapping
    :type system_requirements: dict
    :param system_requirements_by_executable: System requirement by executable double mapping
    :type system_requirements_by_executable: dict
    :param nvidia_driver: a dict mapping function names to nvidia driver requests
    :type nvidia_driver: dict
    :rtype: :class:`~dxpy.bindings.dxjob.DXJob`

    Creates and enqueues a new job that will execute a particular
    function (from the same app or applet as the one the current job is
    running). Returns the :class:`~dxpy.bindings.dxjob.DXJob` handle for
    the job.

    Note that this function is shorthand for::

        dxjob = DXJob()
        dxjob.new(fn_input, fn_name, **kwargs)

    .. note:: This method is intended for calls made from within
       already-executing jobs or apps. If it is called from outside of
       an Execution Environment, an exception will be thrown. To create
       new jobs from outside the Execution Environment, use
       :func:`dxpy.bindings.dxapplet.DXApplet.run` or
       :func:`dxpy.bindings.dxapp.DXApp.run`.

    .. note:: If the environment variable ``DX_JOB_ID`` is not set, this method assmes that it is running within the debug harness, executes the job in place, and provides a debug job handler object that does not have a corresponding remote API job object.

    '''
    dxjob = DXJob()
    dxjob.new(fn_input, fn_name, name=name, tags=tags, properties=properties, details=details,
              instance_type=instance_type, depends_on=depends_on, cluster_spec=cluster_spec, fpga_driver=fpga_driver,
              system_requirements=system_requirements, system_requirements_by_executable=system_requirements_by_executable,
              nvidia_driver=nvidia_driver, **kwargs)
    return dxjob


class DXJob(DXObject):
    '''
    Remote job object handler.
    '''

    _class = "job"

    def __init__(self, dxid=None):
        self._test_harness_result = None
        DXObject.__init__(self, dxid=dxid)
        self.set_id(dxid)

    def new(self, fn_input, fn_name, name=None, tags=None, properties=None, details=None, instance_type=None,
            depends_on=None, cluster_spec=None, fpga_driver=None, system_requirements=None,
            system_requirements_by_executable=None, nvidia_driver=None, **kwargs):
        '''
        :param fn_input: Function input
        :type fn_input: dict
        :param fn_name: Name of the function to be called
        :type fn_name: string
        :param name: Name for the new job (default is "<parent job name>:<fn_name>")
        :type name: string
        :param tags: Tags to associate with the job
        :type tags: list of strings
        :param properties: Properties to associate with the job
        :type properties: dict with string values
        :param details: Details to set for the job
        :type details: dict or list
        :param instance_type: Instance type on which the job will be run, or a dict mapping function names to instance type requests
        :type instance_type: string or dict
        :param depends_on: List of data objects or jobs to wait that need to enter the "closed" or "done" states, respectively, before the new job will be run; each element in the list can either be a dxpy handler or a string ID
        :type depends_on: list
        :param cluster_spec: a dict mapping function names to cluster spec requests
        :type cluster_spec: dict
        :param fpga_driver: a dict mapping function names to fpga driver requests
        :type fpga_driver: dict
        :param system_requirements: System requirement single mapping
        :type system_requirements: dict
        :param system_requirements_by_executable: System requirement by executable double mapping
        :type system_requirements_by_executable: dict
        :param nvidia_driver: a dict mapping function names to nvidia driver requests
        :type nvidia_driver: dict

        Creates and enqueues a new job that will execute a particular
        function (from the same app or applet as the one the current job
        is running).

        .. note:: This method is intended for calls made from within
           already-executing jobs or apps. If it is called from outside
           of an Execution Environment, an exception will be thrown. To
           create new jobs from outside the Execution Environment, use
           :func:`dxpy.bindings.dxapplet.DXApplet.run` or
           :func:`dxpy.bindings.dxapp.DXApp.run`.

        '''
        final_depends_on = []
        if depends_on is not None:
            if isinstance(depends_on, list):
                for item in depends_on:
                    if isinstance(item, DXJob) or isinstance(item, DXDataObject):
                        if item.get_id() is None:
                            raise DXError('A dxpy handler given in depends_on does not have an ID set')
                        final_depends_on.append(item.get_id())
                    elif isinstance(item, basestring):
                        final_depends_on.append(item)
                    else:
                        raise DXError('Expected elements of depends_on to only be either instances of DXJob or DXDataObject, or strings')
            else:
                raise DXError('Expected depends_on field to be a list')

        if 'DX_JOB_ID' in os.environ:
            req_input = {}
            req_input["input"] = fn_input
            req_input["function"] = fn_name
            if name is not None:
                req_input["name"] = name
            if tags is not None:
                req_input["tags"] = tags
            if properties is not None:
                req_input["properties"] = properties
            if any(requirement is not None for requirement in [instance_type, cluster_spec, fpga_driver, nvidia_driver]):
                instance_type_srd = SystemRequirementsDict.from_instance_type(instance_type, fn_name)
                cluster_spec_srd = SystemRequirementsDict(cluster_spec)
                fpga_driver_srd = SystemRequirementsDict(fpga_driver)
                nvidia_driver_srd = SystemRequirementsDict(nvidia_driver)
                req_input["systemRequirements"] = (instance_type_srd + cluster_spec_srd + fpga_driver_srd + nvidia_driver_srd).as_dict()
            if system_requirements is not None:
                req_input["systemRequirements"] = system_requirements
            if system_requirements_by_executable is not None:
                req_input["systemRequirementsByExecutable"] = system_requirements_by_executable
            if depends_on is not None:
                req_input["dependsOn"] = final_depends_on
            if details is not None:
                req_input["details"] = details
            resp = dxpy.api.job_new(req_input, **kwargs)
            self.set_id(resp["id"])
        else:
            self.set_id(queue_entry_point(function=fn_name, input_hash=fn_input,
                                          depends_on=final_depends_on,
                                          name=name))

    def set_id(self, dxid):
        '''
        :param dxid: New job ID to be associated with the handler (localjob IDs also accepted for local runs)
        :type dxid: string

        Discards the currently stored ID and associates the handler with *dxid*
        '''
        if dxid is not None:
            if not (isinstance(dxid, basestring) and dxid.startswith('localjob-')):
                # localjob IDs (which do not follow the usual ID
                # syntax) should be allowed; otherwise, follow the
                # usual syntax checking
                verify_string_dxid(dxid, self._class)
        self._dxid = dxid

    def describe(self, fields=None, defaultFields=None, io=None, **kwargs):
        """
        :param fields: dict where the keys are field names that should
            be returned, and values should be set to True (by default,
            all fields are returned)
        :type fields: dict
        :param defaultFields: include default fields when fields is supplied
        :type defaultFields: bool
        :param io: Include input and output fields in description;
            cannot be provided with *fields*; default is True if
            *fields* is not provided (deprecated)
        :type io: bool
        :returns: Description of the job
        :rtype: dict

        Returns a hash with key-value pairs containing information about
        the job, including its state and (optionally) its inputs and
        outputs, as described in the API documentation for the
        `/job-xxxx/describe
        <https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-describe>`_
        method.

        """
        if fields is not None and io is not None:
            raise DXError('DXJob.describe: cannot provide non-None values for both fields and io')
        describe_input = {}
        if fields is not None:
            describe_input['fields'] = fields
        if defaultFields is not None:
            describe_input['defaultFields'] = defaultFields
        if io is not None:
            describe_input['io'] = io
        self._desc = dxpy.api.job_describe(self._dxid, describe_input, **kwargs)
        return self._desc

    def add_tags(self, tags, **kwargs):
        """
        :param tags: Tags to add to the job
        :type tags: list of strings

        Adds each of the specified tags to the job. Takes no
        action for tags that are already listed for the job.

        """

        dxpy.api.job_add_tags(self._dxid, {"tags": tags}, **kwargs)

    def remove_tags(self, tags, **kwargs):
        """
        :param tags: Tags to remove from the job
        :type tags: list of strings

        Removes each of the specified tags from the job. Takes
        no action for tags that the job does not currently have.

        """

        dxpy.api.job_remove_tags(self._dxid, {"tags": tags}, **kwargs)

    def update(self, allow_ssh, **kwargs):
        """
        :param allow_ssh: Allowable IP ranges to set for SSH access to the job
        :type allow_ssh: list of strings

        Updates a job's allowSSH field, overwrites existing values

        """

        dxpy.api.job_update(self._dxid, {"allowSSH": allow_ssh}, **kwargs)


    def set_properties(self, properties, **kwargs):
        """
        :param properties: Property names and values given as key-value pairs of strings
        :type properties: dict

        Given key-value pairs in *properties* for property names and
        values, the properties are set on the job for the given
        property names. Any property with a value of :const:`None`
        indicates the property will be deleted.

        .. note:: Any existing properties not mentioned in *properties*
           are not modified by this method.

        """

        dxpy.api.job_set_properties(self._dxid, {"properties": properties}, **kwargs)

    def wait_on_done(self, interval=2, timeout=3600*24*7, **kwargs):
        '''
        :param interval: Number of seconds between queries to the job's state
        :type interval: integer
        :param timeout: Maximum amount of time to wait, in seconds, until the job is done running
        :type timeout: integer
        :raises: :exc:`~dxpy.exceptions.DXError` if the timeout is reached before the job has finished running, or :exc:`dxpy.exceptions.DXJobFailureError` if the job fails

        Waits until the job has finished running.
        '''

        elapsed = 0
        while True:
            state = self._get_state(**kwargs)
            if state == "done":
                break
            if state == "failed":
                desc = self.describe(**kwargs)
                err_msg = "Job has failed because of {failureReason}: {failureMessage}".format(**desc)
                if desc.get("failureFrom") != None and desc["failureFrom"]["id"] != desc["id"]:
                    err_msg += " (failure from {id})".format(id=desc['failureFrom']['id'])
                raise DXJobFailureError(err_msg)
            if state == "terminated":
                raise DXJobFailureError("Job was terminated.")

            if elapsed >= timeout or elapsed < 0:
                raise DXJobFailureError("Reached timeout while waiting for the job to finish")

            time.sleep(interval)
            elapsed += interval

    def terminate(self, **kwargs):
        '''
        Terminates the associated job.
        '''
        dxpy.api.job_terminate(self._dxid, **kwargs)

    def get_output_ref(self, field, index=None, metadata=None):
        '''
        :param field: Output field name of this job
        :type field: string
        :param index: If the referenced field is an array, optionally specify an index (starting from 0) to indicate a particular member of the array
        :type index: int
        :param metadata: If the referenced field is of a data object class, a string indicating the metadata that should be read, e.g. "name", "properties.propkey", "details.refgenome"
        :type metadata: string

        Returns a dict containing a valid job-based object reference
        to refer to an output of this job.  This can be used directly
        in place of a DNAnexus link when used as a job output value.
        For example, after creating a subjob, the following app
        snippet uses a reference to the new job's output as part of
        its own output::

            mysubjob = dxpy.new_dxjob({}, "my_function")
            return { "myfileoutput": mysubjob.get_output_ref("output_field_name"),
                     "myotherfileoutput": mysubjob.get_output_ref("output_array",
                                                                  index=1),
                     "filename": mysubjob.get_output_ref("output_field_name",
                                                         metadata="name") }
        '''

        link = {"$dnanexus_link": {"job": self._dxid, "field": field}}
        if index is not None:
            link["$dnanexus_link"]["index"] = index
        if metadata is not None:
            link["$dnanexus_link"]["metadata"] = metadata
        return link

    def _get_state(self, **kwargs):
        '''
        :returns: State of the remote object
        :rtype: string

        Queries the API server for the job's state.

        Note that this function is shorthand for:

            dxjob.describe(io=False, **kwargs)["state"]

        '''

        return self.describe(fields=dict(state=True), **kwargs)["state"]



import sys
import gzip
from dxpy import DXHTTPRequest
from dxpy.bindings import DXRecord, DXFile
import json


class Dataset(DXRecord):
    def __init__(self, dataset_id, detail_describe_dict=None):
        super(DXRecord, self).__init__(dataset_id)
        self.dataset_id = dataset_id
        self._detail_describe = detail_describe_dict
        self._visualize_info = None

        if detail_describe_dict:
            if "details" not in detail_describe_dict:
                raise ValueError("detail is expected key in detail_describe_dict")

    @staticmethod
    def resolve_cohort_to_dataset(record_obj):
        record_obj_desc = record_obj.describe(
            default_fields=True, fields={"properties", "details"}
        )
        cohort_info = None
        is_cohort = "CohortBrowser" in record_obj_desc["types"]

        if is_cohort:
            cohort_info = record_obj_desc
            dataset_id = record_obj_desc["details"]["dataset"]["$dnanexus_link"]
            dataset_obj = Dataset(dataset_id)
        else:
            dataset_obj = Dataset(record_obj.id, record_obj_desc)

        return dataset_obj, cohort_info

    @property
    def descriptor_file(self):
        return self.detail_describe["details"]["descriptor"]["$dnanexus_link"]

    @property
    def descriptor_file_dict(self):
        is_python2 = sys.version_info.major == 2
        content = DXFile(
            self.descriptor_file, mode="rb", project=self.project_id
        ).read()

        if is_python2:
            import StringIO

            x = StringIO.StringIO(content)
            file_obj = gzip.GzipFile(fileobj=x)
            content = file_obj.read()

        else:
            content = gzip.decompress(content)

        return json.loads(content)

    @property
    def visualize_info(self):
        if self._visualize_info is None:
            self._visualize_info = DXHTTPRequest(
                "/" + self.dataset_id + "/visualize",
                {"project": self.project_id, "cohortBrowser": False},
            )
        return self._visualize_info

    @property
    def vizserver_url(self):
        vis_info = self.visualize_info
        return vis_info.get("url")

    @property
    def project_id(self):
        return self.detail_describe.get("project")

    @property
    def version(self):
        return self.detail_describe.get("details").get("version")

    @property
    def detail_describe(self):
        if self._detail_describe is None:
            self._detail_describe = self.describe(
                default_fields=True, fields={"properties", "details"}
            )
        return self._detail_describe

    @property
    def assays_info_dict(self):
        assays = self.descriptor_file_dict["assays"]
        assays_info_dict = {}

        for index in range(len(assays)):
            model = assays[index]["generalized_assay_model"]
            assay_dict = {
                "name": assays[index]["name"],
                "index": index,
                "uuid": assays[index]["uuid"],
                "reference": assays[index].get("reference"),
                "generalized_assay_model_version": assays[index].get(
                    "generalized_assay_model_version"
                ),
            }

            if model not in assays_info_dict.keys():
                assays_info_dict[model] = []

            assays_info_dict[model].append(assay_dict)

        return assays_info_dict

    def assay_info_dict(self, assay_uuid):
        for assay_list in self.assays_info_dict.values():
            for assay in assay_list:
                if assay["uuid"] == assay_uuid:
                    return assay

    def assay_names_list(self, assay_type):
        assay_names_list = []
        if self.assays_info_dict.get(assay_type):
            for assay in self.assays_info_dict.get(assay_type):
                assay_names_list.append(assay["name"])
        return assay_names_list

    def is_assay_name_valid(self, assay_name, assay_type):
        return assay_name in self.assay_names_list(assay_type)

    def assay_index(self, assay_name):
        assay_lists_per_model = self.assays_info_dict.values()
        for model_assays in assay_lists_per_model:
            for assay in model_assays:
                if assay["name"] == assay_name:
                    return assay["index"]



# Copyright (C) 2014-2020 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import print_function
import subprocess
import json
import os
from dxpy.utils import file_load_utils
from dxpy.exceptions import err_exit

def _build_mount_manifest(to_mount):
    files_list = []
    for file_rec in to_mount:
        file_entry = {}
        file_handler = file_rec['handler']
        file_entry['proj_id'] = file_handler.get_proj_id()
        file_entry['file_id'] = file_rec['src_file_id']
        file_name = file_rec['trg_fname']
        file_entry['parent'] = "/" + os.path.dirname(file_name)
        files_list.append(file_entry)
    files_manifest = {'Files': files_list}
    print("File mount manifest:")
    print(files_manifest)
    return files_manifest

def _which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    for path in os.environ["PATH"].split(os.pathsep):
        exe_file = os.path.join(path, program)
        if is_exe(exe_file):
            return exe_file
    return None

def _gen_helper_dict(filtered_inputs):
    '''
    Create a dict of values for the mounted files. This is similar to the variables created
    when running a bash app.
    '''

    file_key_descs, _ignore = file_load_utils.analyze_bash_vars(
        file_load_utils.get_input_json_file(), None)

    flattened_dict = {}

    def add_if_no_collision(key, value, dict_):
        if key not in dict_:
            dict_[key] = value

    for input_ in filtered_inputs:
        if input_ not in file_key_descs:
            continue
        input_var_dict = file_key_descs[input_]
        add_if_no_collision(input_ + '_path', input_var_dict["path"], flattened_dict)
        add_if_no_collision(input_ + '_name', input_var_dict["basename"], flattened_dict)
        add_if_no_collision(input_ + '_prefix', input_var_dict["prefix"], flattened_dict)

    return flattened_dict


def mount_all_inputs(exclude=None, verbose=False):
    '''
    :param exclude: List of input variables that should not be mounted.
    :type exclude: Array of strings
    :returns: dict of lists of strings where each key is the input variable
                and each list element is the full path to the file that has
                been mounted.
    :param verbose: Start dxfuse with '-verbose 2' logging
    :type verbose: boolean


    This function mounts all files that were supplied as inputs to the app.
    By convention, if an input parameter "FOO" has value

        {"$dnanexus_link": "file-xxxx"}

    and filename INPUT.TXT, then the linked file will be mounted into the
    path:

        $HOME/in/FOO/INPUT.TXT

    If an input is an array of files, then all files will be placed into
    numbered subdirectories under a parent directory named for the
    input. For example, if the input key is FOO, and the inputs are {A, B,
    C}.vcf then, the directory structure will be:

        $HOME/in/FOO/0/A.vcf
                     1/B.vcf
                     2/C.vcf

    Zero padding is used to ensure argument order. For example, if there are
    12 input files {A, B, C, D, E, F, G, H, I, J, K, L}.txt, the directory
    structure will be:

        $HOME/in/FOO/00/A.vcf
                     ...
                     11/L.vcf

    This allows using shell globbing (FOO/*/*.vcf) to get all the files in the input
    order and prevents issues with files which have the same filename.'''

    print("Mounting inputs...")

    home_dir = os.environ["HOME"]
    mount_dir = os.path.join(home_dir, "in")
    mount_manifest_file = os.path.join(home_dir, "mount-manifest.json")
    dxfuse_cmd = _which("dxfuse")
    if dxfuse_cmd is None:
        err_exit("dxfuse is not installed on this system")

    subprocess.check_output(["mkdir", mount_dir])

    try:
        job_input_file = file_load_utils.get_input_json_file()
        dirs, inputs, rest = file_load_utils.get_job_input_filenames(job_input_file)
    except IOError:
        msg = 'Error: Could not find the input json file: {0}.\n'.format(job_input_file)
        msg += '       This function should only be called from within a running job.'
        print(msg)
        raise

    # Remove excluded inputs
    if exclude:
        inputs = file_load_utils.filter_dict(inputs, exclude)

    # Convert to a flat list of elements to mount
    to_mount = []
    for ival_list in inputs.values():
        to_mount.extend(ival_list)

    files_manifest = _build_mount_manifest(to_mount)
    with open(mount_manifest_file, 'w') as mfile:
        json.dump(files_manifest, mfile)

    dxfuse_version = subprocess.check_output([dxfuse_cmd, "-version"])
    print("Using dxfuse version " + str(dxfuse_version))

    uid = str(int(subprocess.check_output(["id", "-u"])))
    gid = str(int(subprocess.check_output(["id", "-g"])))
    cmd = [dxfuse_cmd, "-uid", uid, "-gid", gid, mount_dir, mount_manifest_file]
    if verbose:
        cmd[1:1] = ["-verbose", "2"]
    print(subprocess.check_output(cmd))

    print("Done mounting inputs.")

    subprocess.call(["find", mount_dir, "-name", "*"])

    helper_vars = _gen_helper_dict(inputs)
    return helper_vars



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
This file contains parsers with no added help that can be inherited by
other parsers, as well as utility functions for parsing the input to
those parsers.
'''

from __future__ import print_function, unicode_literals, division, absolute_import

import argparse, json
from .. import config
from ..utils.printing import fill
from ..utils.pretty_print import format_table
from ..utils.resolver import split_unescaped
from ..utils.completer import InstanceTypesCompleter
from ..exceptions import (DXError, DXCLIError)
from ..compat import basestring

class DXParserError(DXError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

all_arg = argparse.ArgumentParser(add_help=False)
all_arg.add_argument('-a', '--all', help=fill('Apply to all results with the same name without prompting', width_adjustment=-24), action='store_true')

no_color_arg = argparse.ArgumentParser(add_help=False)
no_color_arg.add_argument('--color',
                          help=fill('Set when color is used (color=auto is used when stdout is a TTY)', width_adjustment=-24),
                          choices=['off', 'on', 'auto'], default='auto')

delim_arg = argparse.ArgumentParser(add_help=False)
delim_arg.add_argument('--delimiter', '--delim',
                       dest='delimiter',
                       help=fill('Always use exactly one of DELIMITER to separate fields to be printed; if no delimiter is provided with this flag, TAB will be used', width_adjustment=-24),
                       nargs='?',
                       const="\t")

json_arg = argparse.ArgumentParser(add_help=False)
json_arg.add_argument('--json', help='Display return value in JSON', action='store_true')

try_arg = argparse.ArgumentParser(add_help=False)
try_arg.add_argument('--try', metavar="T", dest="job_try", type=int,
                     help=fill('When modifying a job that was restarted, apply the change to try T of the restarted job. T=0 refers to the first try. Default is the last job try.', width_adjustment=-24))

stdout_args = argparse.ArgumentParser(add_help=False)
stdout_args_gp = stdout_args.add_mutually_exclusive_group()
stdout_args_gp.add_argument('--brief', help=fill('Display a brief version of the return value; for most commands, prints a DNAnexus ID per line', width_adjustment=-24), action='store_true')
stdout_args_gp.add_argument('--verbose', help='If available, displays extra verbose output',
                            action='store_true')

parser_dataobject_args = argparse.ArgumentParser(add_help=False)
parser_dataobject_args_gp = parser_dataobject_args.add_argument_group('metadata arguments')
parser_dataobject_args_gp.add_argument('--visibility', choices=['hidden', 'visible'], dest='hidden', default='visible', help='Whether the object is hidden or not')
parser_dataobject_args_gp.add_argument('--property', dest='properties', metavar='KEY=VALUE', help=fill('Key-value pair to add as a property; repeat as necessary,', width_adjustment=-24) + '\n' + fill('e.g. "--property key1=val1 --property key2=val2"', width_adjustment=-24, initial_indent=' ', subsequent_indent=' ', break_on_hyphens=False), action='append')
parser_dataobject_args_gp.add_argument('--type', metavar='TYPE', dest='types', help=fill('Type of the data object; repeat as necessary,', width_adjustment=-24) + '\n' + fill('e.g. "--type type1 --type type2"', width_adjustment=-24, break_on_hyphens=False, initial_indent=' ', subsequent_indent=' '), action='append')
parser_dataobject_args_gp.add_argument('--tag', metavar='TAG', dest='tags', help=fill('Tag of the data object; repeat as necessary,', width_adjustment=-24) + '\n' + fill('e.g. "--tag tag1 --tag tag2"', width_adjustment=-24, break_on_hyphens=False, initial_indent=' ', subsequent_indent=' '), action='append')
parser_dataobject_args_gp.add_argument('--details', help='JSON to store as details')
parser_dataobject_args_gp.add_argument('-p', '--parents', help='Create any parent folders necessary', action='store_true')

parser_single_dataobject_output_args = argparse.ArgumentParser(add_help=False)
parser_single_dataobject_output_args.add_argument('-o', '--output', help=argparse.SUPPRESS)
parser_single_dataobject_output_args.add_argument('path', help=fill('DNAnexus path for the new data object (default uses current project and folder if not provided)', width_adjustment=-24), nargs='?')

find_by_properties_and_tags_args = argparse.ArgumentParser(add_help=False)
find_by_properties_and_tags_args.add_argument('--property', dest='properties',
                                              metavar='KEY[=VALUE]',
                                              help=fill('Key-value pair of a property or simply a property key; if only a key is provided, matches a result that has the key with any value; repeat as necessary, e.g. "--property key1=val1 --property key2"', width_adjustment=-24),
                                              action='append')
find_by_properties_and_tags_args.add_argument('--tag',
                                              help=fill('Tag to match; repeat as necessary, e.g. "--tag tag1 --tag tag2" will require both tags', width_adjustment=-24),
                                              action='append')

find_executions_args = argparse.ArgumentParser()
find_executions_args.add_argument('--id', help=fill('Show only the job tree or job containing this job ID', width_adjustment=-24))
find_executions_args.add_argument('--name', help=fill('Restrict the search by job name (accepts wildcards "*" and "?")', width_adjustment=-24))
find_executions_args.add_argument('--user', help=fill('Username who launched the job (use "self" to ask for your own jobs)', width_adjustment=-24))
find_executions_args.add_argument('--project', help=fill('Project context (output project), default is current project if set', width_adjustment=-24))
find_executions_args.add_argument('--all-projects', '--allprojects', help=fill('Extend search to all projects', width_adjustment=-24), action='store_true')
find_executions_args.add_argument('--app', '--applet', '--executable', dest='executable', help=fill('Applet or App ID that job is running', width_adjustment=-24))
find_executions_args.add_argument('--state', help=fill('State of the job, e.g. \"done\", \"failed\"', width_adjustment=-24))
find_executions_args.add_argument('--origin', help=fill('Job ID of the top-level job', width_adjustment=-24)) # Redundant but might as well
find_executions_args.add_argument('--parent', help=fill('Job ID of the parent job; implies --all-jobs', width_adjustment=-24))
find_executions_args.add_argument('--created-after', help=fill('Date (e.g. --created-after="2021-12-01" or --created-after="2021-12-01 19:01:33") or integer Unix epoch timestamp in milliseconds (e.g. --created-after=1642196636000) after which the job was last created. You can also specify negative numbers to indicate a time period in the past suffixed by s, m, h, d, w, M or y to indicate seconds, minutes, hours, days, weeks, months or years (e.g. --created-after=-2d for executions created in the last 2 days)', width_adjustment=-24))
find_executions_args.add_argument('--created-before', help=fill('Date (e.g. --created-before="2021-12-01" or --created-before="2021-12-01 19:01:33") or integer Unix epoch timestamp in milliseconds (e.g. --created-before=1642196636000) before which the job was last created. You can also specify negative numbers to indicate a time period in the past suffixed by s, m, h, d, w, M or y to indicate seconds, minutes, hours, days, weeks, months or years (e.g. --created-before=-2d for executions created earlier than 2 days ago)', width_adjustment=-24))
find_executions_args.add_argument('--no-subjobs', help=fill('Do not show any subjobs', width_adjustment=-24), action='store_true')
find_executions_args.add_argument('--root-execution', '--root', help=fill('Execution ID of the top-level (user-initiated) job or analysis', width_adjustment=-24))
find_executions_args.add_argument('-n', '--num-results', metavar='N', type=int, help=fill('Max number of results (trees or jobs, as according to the search mode) to return (default 10)', width_adjustment=-24), default=10)
find_executions_args.add_argument('-o', '--show-outputs', help=fill('Show job outputs in results', width_adjustment=-24), action='store_true')
find_executions_args.add_argument('--include-restarted', help=fill('if specified, results will include restarted jobs and job trees rooted in restarted jobs', width_adjustment=-24), action='store_true')

def add_find_executions_search_gp(parser):
    find_executions_search_gp = parser.add_argument_group('Search mode')
    find_executions_search = find_executions_search_gp.add_mutually_exclusive_group()
    find_executions_search.add_argument('--trees', help=fill('Show entire job trees for all matching results (default)', width_adjustment=-24), action='store_true')
    find_executions_search.add_argument('--origin-jobs', help=fill('Search and display only top-level origin jobs', width_adjustment=-24), action='store_true')
    find_executions_search.add_argument('--all-jobs', help=fill('Search for jobs at all depths matching the query (no tree structure shown)', width_adjustment=-24), action='store_true')

def process_properties_args(args):
    # Properties
    properties = None
    if args.properties is not None:
        properties = {}
        for keyeqval in args.properties:
            substrings = split_unescaped('=', keyeqval, include_empty_strings=True)
            if len(substrings) != 2:
                raise DXParserError('Property key-value pair must be given using syntax "property_key=property_value"')
            elif substrings[0] == '':
                raise DXParserError('Property keys must be nonempty strings')
            else:
                properties[substrings[0]] = substrings[1]
    args.properties = properties

def process_find_by_property_args(args):
    properties = None
    if args.properties is not None:
        properties = {}
        for keyeqval in args.properties:
            substrings = split_unescaped('=', keyeqval, include_empty_strings=True)
            if len(substrings) > 2:
                raise DXParserError('Property value must be given using syntax "property_key" or "property_key=property_value"')
            elif substrings[0] == '':
                raise DXParserError('Property keys must be nonempty strings')
            elif len(substrings) == 1:
                properties[keyeqval] = True
            else:
                properties[substrings[0]] = substrings[1]

    args.properties = properties

def process_dataobject_args(args):
    process_properties_args(args)

    # Visibility
    args.hidden = (args.hidden == 'hidden' or args.hidden is True)

    # Details
    if args.details is not None:
        try:
            args.details = json.loads(args.details)
        except:
            raise DXParserError('Error: details could not be parsed as JSON')

def process_single_dataobject_output_args(args):
    if args.path is not None and args.output is not None:
        raise DXParserError('Error: Cannot provide both the positional PATH and -o/--output arguments')
    elif args.output is None:
        args.output = args.path

_env_args = argparse.ArgumentParser(add_help=False, prog='dx command ...')
_env_args.add_argument('--apiserver-host', help='API server host')
_env_args.add_argument('--apiserver-port', help='API server port')
_env_args.add_argument('--apiserver-protocol', help='API server protocol (http or https)')
_env_args.add_argument('--project-context-id', help='Default project or project context ID')
_env_args.add_argument('--workspace-id', help='Workspace ID (for jobs only)')
_env_args.add_argument('--security-context', help='JSON string of security context')
_env_args.add_argument('--auth-token', help='Authentication token')

class EnvHelpAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        _env_args.print_help()
        parser.exit(0)

env_args = argparse.ArgumentParser(add_help=False)
env_args.add_argument('--apiserver-host', help=argparse.SUPPRESS)
env_args.add_argument('--apiserver-port', help=argparse.SUPPRESS)
env_args.add_argument('--apiserver-protocol', help=argparse.SUPPRESS)
env_args.add_argument('--project-context-id', help=argparse.SUPPRESS)
env_args.add_argument('--workspace-id', help=argparse.SUPPRESS)
env_args.add_argument('--security-context', help=argparse.SUPPRESS)
env_args.add_argument('--auth-token', help=argparse.SUPPRESS)
env_args.add_argument('--env-help', help=fill('Display help message for overriding environment variables', width_adjustment=-24), action=EnvHelpAction, nargs=0)

def set_env_from_args(args):
    ''' Sets the environment variables for this process from arguments (argparse.Namespace)
    and calls dxpy._initialize() to reset any values that it has already set.
    '''
    args = vars(args)

    if args.get('apiserver_host') is not None:
        config['DX_APISERVER_HOST'] = args['apiserver_host']
    if args.get('apiserver_port') is not None:
        config['DX_APISERVER_PORT'] = args['apiserver_port']
    if args.get('apiserver_protocol') is not None:
        config['DX_APISERVER_PROTOCOL'] = args['apiserver_protocol']
    if args.get('project_context_id') is not None:
        config['DX_PROJECT_CONTEXT_ID'] = args['project_context_id']
    if args.get('workspace_id') is not None:
        config['DX_WORKSPACE_ID'] = args['workspace_id']
    if args.get('cli_wd') is not None:
        config['DX_CLI_WD'] = args['cli_wd']
    if args.get('security_context') is not None:
        config['DX_SECURITY_CONTEXT'] = args['security_context']
    if args.get('auth_token') is not None:
        config['DX_SECURITY_CONTEXT'] = json.dumps({"auth_token": args['auth_token'],
                                                    "auth_token_type": "Bearer"})

extra_args = argparse.ArgumentParser(add_help=False)
extra_args.add_argument('--extra-args', help=fill("Arguments (in JSON format) to pass to the underlying API method, overriding the default settings", width_adjustment=-24))

def process_extra_args(args):
    if args.extra_args is not None:
        try:
            args.extra_args = json.loads(args.extra_args)
        except:
            raise DXParserError('Value given for --extra-args could not be parsed as JSON')

exec_input_args = argparse.ArgumentParser(add_help=False)
exec_input_args.add_argument('-i', '--input', help=fill('An input to be added using "<input name>[:<class>]=<input value>" (provide "class" if there is no input spec; it can be any job IO class, e.g. "string", "array:string", or "array"; if "class" is "array" or not specified, the value will be attempted to be parsed as JSON and is otherwise treated as a string)', width_adjustment=-24), action='append')
exec_input_args.add_argument('-j', '--input-json', help=fill('The full input JSON (keys=input field names, values=input field values)', width_adjustment=-24))
exec_input_args.add_argument('-f', '--input-json-file', dest='filename', help=fill('Load input JSON from FILENAME ("-" to use stdin)'))

class PrintInstanceTypeHelp(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print("Help: Specifying instance types for " + parser.prog)
        print()
        print(fill('Instance types can be requested with --instance-type-by-executable and ' +
                   '--instance-type arguments, with --instance-type-by-executable specification ' +
                   'taking priority over --instance-type, workflow\'s stageSystemRequirements, ' +
                   'and specifications provided during app and applet creation.'))
        print()
        print(fill('--instance-type specifications do not propagate to subjobs and sub-analyses ' +
                   'launched from a job with a /executable-xxxx/run call, but --instance-type-by-executable do ' +
                   '(where executable refers to an app, applet or workflow).'))
        print()
        print(fill('When running an app or an applet, --instance-type lets you specify the ' +
                    'instance type to be used by each entry point.'))
        print(fill('A single instance type can be requested to be used by all entry points by providing the instance type name.  Different instance types can also be requested for different entry points of an app or applet by providing a JSON string mapping from function names to instance types, e.g.'))
        print()
        print('    {"main": "mem2_hdd2_v2_x2", "other_function": "mem1_ssd1_v2_x2"}')
        if parser.prog == 'dx run':
            print()
            print(fill('When running a workflow, --instance-type lets you specify instance types for ' +
                       'each entry point of each workflow stage by prepending the request with "<stage identifier>=" ' +
                       '(where a stage identifier is an ID, a numeric index, or a unique stage name) and ' +
                       'repeating the argument for as many stages as desired.  If no stage ' +
                       'identifier is provided, the value is applied as a default for all stages.'))
            print()
            print('Examples')
            print()
            print(fill('1. Run the main entry point of applet-xxxx on mem1_ssd1_v2_x2, and '
                        'all other entry points on mem1_ssd1_v2_x4'))
            print('    dx run applet-xxxx --instance-type \'{"main": "mem1_ssd1_v2_x2",\n' +
                  '                                         "*":    "mem1_ssd1_v2_x4"}\'')
            print()
            print(fill('2. Runs all entry points of the first stage with ' +
                       'mem2_hdd2_v2_x2, the main entry point of the second stage with mem1_ssd1_v2_x4, ' +
                       'the stage named "BWA" with mem1_ssd1_v2_x2, and all other ' +
                       'stages with mem2_hdd2_v2_x4'))
            print()
            print('    dx run workflow-xxxx \\\n' +
                  '     --instance-type 0=mem2_hdd2_v2_x2 \\\n' +
                  '     --instance-type 1=\'{"main": "mem1_ssd1_v2_x4"}\' \\\n' +
                  '     --instance-type BWA=mem1_ssd1_v2_x2 \\\n' +
                  '     --instance-type mem2_hdd2_v2_x4')
            print()
            print(fill('--instance-type-by-executable argument is a JSON string with a double mapping that ' +
                       'specifies instance types by app or applet id, then by entry point within the executable.' +
                       'This specification applies across the entire nested execution tree and is propagated ' +
                       'across /executable-xxxx/run calls issued with the execution tree.'))
            print()
            print('More examples')
            print(fill('3. Force every job in the execution tree to use mem2_ssd1_v2_x2'))
            print()
            print(
                '    dx run workflow-xxxx --instance-type-by-executable \'{"*": {"*": "mem2_ssd1_v2_x2"}}\'')
            print()
            print(fill(
                '4. Force every job in the execution tree executing applet-xyz1 to use mem2_ssd1_v2_x2'))
            print()
            print(
                '    dx run workflow-xxxx --instance-type-by-executable \'{"applet-xyz1":{"*": "mem2_ssd1_v2_x2"}}\'')
            print()
            print(fill('5. Force every job executing applet-xyz1 to use mem2_ssd1_v2_x4 ' +
                       'for the main entry point and mem2_ssd1_v2_x2 for all other entry points.' +
                       'Also force the collect entry point of all executables other than applet-xyz1 to use mem2_ssd1_v2_x8.' +
                       'Other entry points of executable other than applet-xyz1 may be overridden by ' +
                       'lower-priority mechanisms'))
            print()
            print('    dx run workflow-xxxx --instance-type-by-executable \\\n' +
                  '           \'{"applet-xyz1":  {"main":    "mem2_ssd1_v2_x4", "*": "mem2_ssd1_v2_x2"},\n' +
                  '             "*":            {"collect": "mem2_ssd1_v2_x8"}}\'')
            print()
            print(fill('6. Force every job executing applet-xxxx to use mem2_ssd1_v2_x2 for all entry points ' +
                       'in the entire execution tree. ' +
                       'Also force stage 0 executable to run on mem2_ssd1_v2_x4, unless stage 0 invokes ' +
                       'applet-xxxx, in which case applet-xxxx\'s jobs will use mem2_ssd1_v2_x2 as specified by ' +
                       '--instance-type-by-executable.'))
            print()
            print('    dx run workflow-xxxx \\\n' +
                  '     --instance-type-by-executable  \'{"applet-xxxx": {"*": "mem2_ssd1_v2_x2"}}\' \\\n' +
                  '     --instance-type 0=mem2_ssd1_v2_x4')
            print()
            print(fill(
                'See "Requesting Instance Types" in DNAnexus documentation for more details.'))
        print()
        print('Available instance types:')
        print()
        print(format_table(InstanceTypesCompleter.standard_instance_types.values(),
                           column_names=list(InstanceTypesCompleter.standard_instance_types.values())[0]._fields))
        print()
        print('GPU instance types:')
        print()
        print(format_table(InstanceTypesCompleter.gpu_instance_types.values(),
                           column_names=list(InstanceTypesCompleter.gpu_instance_types.values())[0]._fields))
        print()
        print('FPGA instance types:')
        print()
        print(format_table(InstanceTypesCompleter.fpga_instance_types.values(),
                           column_names=list(InstanceTypesCompleter.fpga_instance_types.values())[0]._fields))
        parser.exit(0)

instance_type_arg = argparse.ArgumentParser(add_help=False)
instance_type_arg.add_argument('--instance-type',
                               metavar='INSTANCE_TYPE_OR_MAPPING',
                               help=fill('''When running an app or applet, the mapping lists executable's entry points or "*" as keys, and instance types to use for these entry points as values.  
When running a workflow, the specified instance types can be prefixed by a stage name or stage index followed by "=" to apply to a specific stage, or apply to all workflow stages without such prefix. 
The instance type corresponding to the "*" key is applied to all entry points not explicitly mentioned in the --instance-type mapping. Specifying a single instance type is equivalent to using it for all entry points, so "--instance-type mem1_ssd1_v2_x2" is same as "--instance-type '{"*":"mem1_ssd1_v2_x2"}'. 
Note that "dx run" calls within the execution subtree may override the values specified at the root of the execution tree.
See dx run --instance-type-help for details.
''', width_adjustment=-24, replace_whitespace=False),
                               action='append').completer = InstanceTypesCompleter()

instance_type_arg.add_argument('--instance-type-by-executable',
                               metavar='DOUBLE_MAPPING',
                               help=fill(
                                   '''Specifies instance types by app or applet id, then by entry point within the executable.
The order of priority for this specification is:
  * --instance-type, systemRequirements and stageSystemRequirements specified at runtime
  * stage's systemRequirements, systemRequirements supplied to /app/new and /applet/new at workflow/app/applet build time
  * systemRequirementsByExecutable specified in downstream executions (if any)
See dx run --instance-type-help for details.
''', width_adjustment=-24, replace_whitespace=False))

instance_type_arg.add_argument('--instance-type-help',
                               nargs=0,
                               help=fill('Print help for specifying instance types'),
                               action=PrintInstanceTypeHelp)

property_args = argparse.ArgumentParser(add_help=False)
property_args.add_argument('--property', dest='properties', metavar='KEY=VALUE',
                           help=(fill('Key-value pair to add as a property; repeat as necessary,',
                                      width_adjustment=-24) + '\n' +
                                 fill('e.g. "--property key1=val1 --property key2=val2"',
                                      width_adjustment=-24, initial_indent=' ', subsequent_indent=' ',
                                      break_on_hyphens=False)),
                           action='append')

tag_args = argparse.ArgumentParser(add_help=False)
tag_args.add_argument('--tag', metavar='TAG', dest='tags',
                      help=fill('Tag for the resulting execution; repeat as necessary,', width_adjustment=-24) + '\n' +
                      fill('e.g. "--tag tag1 --tag tag2"', width_adjustment=-24,
                           break_on_hyphens=False, initial_indent=' ', subsequent_indent=' '),
                      action='append')

contains_phi = argparse.ArgumentParser(add_help=False)
contains_phi.add_argument('--phi', dest='containsPHI', choices=["true", "false"],
                          help='If set to true, only projects that contain PHI data will be retrieved. ' +
                          'If set to false, only projects that do not contain PHI data will be retrieved.')

def _parse_dictionary_or_string_input(thing, arg_name):
    if thing.strip().startswith('{'):
        # expects a map, e.g of entry point to instance type or instance count
        try:
            return json.loads(thing)
        except ValueError:
            raise DXCLIError("Error while parsing JSON value for " + arg_name)
    else:
        return thing

def _parse_inst_type(thing):
    return _parse_dictionary_or_string_input(thing, "--instance-type")

def _parse_inst_count(thing):
    return _parse_dictionary_or_string_input(thing, "--instance-count")

def process_instance_type_arg(args, for_workflow=False):
    if args.instance_type:
        if for_workflow:
            args.stage_instance_types = {}
            new_inst_type_val = {}
            for inst_type_req in args.instance_type:
                if '=' in inst_type_req:
                    index_of_eql = inst_type_req.rfind('=')
                    args.stage_instance_types[inst_type_req[:index_of_eql]] = _parse_inst_type(
                        inst_type_req[index_of_eql + 1:]
                    )
                else:
                    new_inst_type_val = _parse_inst_type(inst_type_req)
            args.instance_type = new_inst_type_val
        elif not isinstance(args.instance_type, basestring):
            args.instance_type = _parse_inst_type(args.instance_type[-1])
        else:
            # is a string
            args.instance_type = _parse_inst_type(args.instance_type)

def process_instance_type_by_executable_arg(args):
    if args.instance_type_by_executable:
        if args.instance_type_by_executable.strip().startswith('{'):
            # expects a map, e.g of entry point to instance type or instance count
            try:
                args.instance_type_by_executable = json.loads(args.instance_type_by_executable)
            except ValueError:
                raise DXParserError('Error while parsing JSON value for --instance-type-by-executable')
        else:
            raise DXParserError('Value given for --instance-type-by-executable could not be parsed as JSON')

def process_instance_count_arg(args):
    if args.instance_count:
        # If --instance-count was used multiple times, the last one
        # from the list will override the previous ones
        if not isinstance(args.instance_count, basestring):
            args.instance_count = _parse_inst_count(args.instance_count[-1])
        else:
            # is a string
            args.instance_count = _parse_inst_count(args.instance_count)

def get_update_project_args(args):
    input_params = {}
    if args.name is not None:
        input_params["name"] = args.name
    if args.summary is not None:
        input_params["summary"] = args.summary
    if args.description is not None:
        input_params["description"] = args.description
    if args.protected is not None:
        input_params["protected"] = args.protected == 'true'
    if args.restricted is not None:
        input_params["restricted"] = args.restricted == 'true'
    if args.download_restricted is not None:
        input_params["downloadRestricted"] = args.download_restricted == 'true'
    if args.containsPHI is not None:
        input_params["containsPHI"] = args.containsPHI == 'true'
    if args.database_ui_view_only is not None:
        input_params["databaseUIViewOnly"] = args.database_ui_view_only == 'true'
    if args.bill_to is not None:
        input_params["billTo"] = args.bill_to
    if args.allowed_executables is not None:
        input_params['allowedExecutables'] = args.allowed_executables
    if args.unset_allowed_executables:
        input_params['allowedExecutables'] = None
    if args.database_results_restricted is not None:
        input_params['databaseResultsRestricted'] = args.database_results_restricted
    if args.unset_database_results_restricted:
        input_params['databaseResultsRestricted'] = None
    if args.external_upload_restricted is not None:
        input_params['externalUploadRestricted'] = args.external_upload_restricted == 'true'
    if args.https_app_isolated_browsing is not None:
        input_params['httpsAppIsolatedBrowsing'] = args.https_app_isolated_browsing == 'true'
    return input_params

def process_phi_param(args):
    if args.containsPHI is not None:
        if args.containsPHI == "true":
            args.containsPHI = True
        elif args.containsPHI == "false":
            args.containsPHI = False

def process_external_upload_restricted_param(args):
    if args.external_upload_restricted is not None:
        args.external_upload_restricted = (args.external_upload_restricted == "true")



import os

from ..exceptions import err_exit
from .retrieve_bins import retrieve_bins
import dxpy

# A dictionary relating the user-facing names of columns to their actual column

column_conversion = {
    "allele_id": "variant_read_optimized$allele_id",
    "variant_type": "variant_read_optimized$variant_type",
    "symbol": "variant_read_optimized$SYMBOL",
    "gene": "variant_read_optimized$Gene",
    "feature": "variant_read_optimized$Feature",
    "hgvsc": "variant_read_optimized$HGVSc",
    "hgvsp": "variant_read_optimized$HGVSp",
    "assay_sample_id": "variant_read_optimized$assay_sample_id",
    "sample_id": "variant_read_optimized$sample_id",
    "tumor_normal": "variant_read_optimized$tumor_normal",
}

column_conditions = {
    "allele_id": "in",
    "variant_type": "in",
    "symbol": "any",
    "gene": "any",
    "feature": "any",
    "hgvsc": "any",
    "hgvsp": "any",
    "assay_sample_id": "in",
    "sample_id": "in",
    "tumor_normal": "is",
}

extract_utils_basepath = os.path.join(
    os.path.dirname(dxpy.__file__), "dx_extract_utils"
)

chrs = []
for i in range(1, 23):
    chrs.append(str(i))
chrs.extend(["Y", "X"])

def retrieve_geno_bins(list_of_genes, project, genome_reference, friendly_name):
    """
    A function for determining appropriate geno bins to attach to a given SYMBOL, Gene or Feature
    """
    stage_file = "Homo_sapiens_genes_manifest_staging_vep.json"
    platform_file = "Homo_sapiens_genes_manifest_vep.json"
    error_message = "Following symbols, genes or features are invalid"
    genome_reference = genome_reference + "_" + friendly_name
    geno_bins = retrieve_bins(list_of_genes, project, genome_reference, extract_utils_basepath,
                  stage_file,platform_file,error_message)
    updated_bins = []
    # If a gene, symbol or feature has non standard contig,
    # The correct bin is 'Other'
    for bin in geno_bins:
        if bin['chr'].strip("chr").strip("Chr") not in chrs:
            bin['chr'] = 'Other'
        updated_bins.append(bin)
    return updated_bins

def basic_filter(
    table, friendly_name, values=[], project_context=None, genome_reference=None
):
    """
    A low-level filter consisting of a dictionary with one key defining the table and column
    and whose value is dictionary defining the user-provided value to be compared to, and the logical operator
    used to do the comparison
    ex.
    {"allele$allele_type": [
        {
            "condition": "in",
            "values": [
                "SNP"
            ]
        }
    ]}
    """
    table = "variant_read_optimized"
    # Get the name of this field in the variant table
    # If the column isn't in the regular fields list, use the friendly name itself as the column name
    # This could be the case when "--additional-fields" flag is used
    filter_key = column_conversion.get(
        friendly_name, "variant_read_optimized${}".format(friendly_name)
    )
    # Handle special cases where values need to be capitalized
    if friendly_name == "variant_type" or friendly_name == "gene" or friendly_name == "feature":
        values = [str(x).upper() for x in values]
    # Get the condition ofr this field
    condition = column_conditions.get(friendly_name, "in")

    # Check if we need to add geno bins as well
    if friendly_name in ["symbol", "gene", "feature"]:
        listed_filter = {
            filter_key: [
                {
                    "condition": condition,
                    "values": values,
                    "geno_bins": retrieve_geno_bins(
                        values, project_context, genome_reference, friendly_name
                    ),
                }
            ]
        }
    else:
        listed_filter = {filter_key: [{"condition": condition, "values": values}]}
    
    return listed_filter

def location_filter(location_list):
    """
    A location filter is actually an variant_read_optimized$allele_id filter with no filter values
    The geno_bins perform the actual location filtering. Related to other geno_bins filters by "or"
    """

    location_aid_filter = {
        "variant_read_optimized$allele_id": [
            {
                "condition": "in",
                "values": [],
                "geno_bins": [],
            }
        ]
    }

    chrom =[]

    for location in location_list:
        # First, ensure that the geno bins width isn't greater than 250 megabases
        start = int(location["starting_position"])
        end = int(location["ending_position"])
        if end - start > 250000000:
            err_exit(
                "Error in location {}\nLocation filters may not specify regions larger than 250 megabases".format(
                    location
                )
            )

        # Fill out the contents of an object in the geno_bins array
        chrom.append(location["chromosome"])
        chr = location["chromosome"].strip("chr").strip("Chr")
        
        # If a non standard contig ID is passed in location filter's chr,
        # Add the filter as a CHROM filter and pass 'Other' in geno_bins
        # Standard contigs are also passed to CHROM filter
        if chr not in chrs:
            chr = "Other"
            
        location_aid_filter["variant_read_optimized$allele_id"][0]["geno_bins"].append(
            {
                "chr": chr,
                "start": start,
                "end": end,
            }
        )
    
    return location_aid_filter, list(set(chrom))


def generate_assay_filter(
    full_input_dict,
    name,
    id,
    project_context,
    genome_reference=None,
    include_normal=False,
):
    """
    Generate asasy filter consisting of a compound that links the Location filters if present
    to the regular filters
    {
        "assay_filters": {
            "id": "<id>",
            "name":"<name>",
            "compound":[
                {
                    <contents of location compound>
                }
            ]
        }
    }
    """

    filters_dict = {}

    for filter_group in full_input_dict.keys():
        if filter_group == "location":
            location_list = full_input_dict["location"]
            if location_list:
                location_aid_filter, chrom = location_filter(location_list)
                filters_dict.update(location_aid_filter)
                # Add a CHROM filter to handle non standard contigs
                filters_dict.update(basic_filter(
                    "variant_read_optimized",
                    "CHROM",
                    chrom,
                    project_context,
                    genome_reference,
                ))
            
        else:
            for individual_filter_name in full_input_dict[filter_group].keys():
                individual_filter = full_input_dict[filter_group][individual_filter_name]
                if individual_filter:
                    filters_dict.update(basic_filter(
                        "variant_read_optimized",
                        individual_filter_name,
                        individual_filter,
                        project_context,
                        genome_reference,
                    ))

    # If include_normal is False, then add a filter to select data where tumor_normal = tumor
    if not include_normal:
        filters_dict.update(basic_filter(
            "variant_read_optimized",
            "tumor_normal",
            "tumor",
            project_context,
            genome_reference,
        ))

    final_filter_dict = {"assay_filters": {"name": name, "id": id}}

    # Additional structure of the payload
    final_filter_dict["assay_filters"].update({"filters": filters_dict})
    # The general filters are related by "and"
    final_filter_dict["assay_filters"]["logic"] = "and"

    return final_filter_dict


def somatic_final_payload(
    full_input_dict,
    name,
    id,
    project_context,
    genome_reference=None,
    additional_fields=None,
    include_normal=False,
):
    """
    Assemble the top level payload.  Top level dict contains the project context, fields (return columns),
    and raw filters objects.  This payload is sent in its entirety to the vizserver via an
    HTTPS POST request
    """
    # Generate the assay filter component of the payload
    assay_filter = generate_assay_filter(
        full_input_dict, name, id, project_context, genome_reference, include_normal
    )

    final_payload = {}
    # Set the project context
    final_payload["project_context"] = project_context

    fields = [
        {"assay_sample_id": "variant_read_optimized$assay_sample_id"},
        {"allele_id": "variant_read_optimized$allele_id"},
        {"CHROM": "variant_read_optimized$CHROM"},
        {"POS": "variant_read_optimized$POS"},
        {"REF": "variant_read_optimized$REF"},
        {"allele": "variant_read_optimized$allele"},
    ]

    order_by = [
        {"CHROM":"asc"},
        {"POS":"asc"},
        {"allele_id":"asc"},
        {"assay_sample_id":"asc"}
    ]

    # If the user has specified additional return columns, add them to the payload here
    if additional_fields:
        for add_field in additional_fields:
            fields.append(
                {"{}".format(add_field): "variant_read_optimized${}".format(add_field)}
            )

    final_payload["fields"] = fields
    final_payload["order_by"] = order_by
    final_payload["raw_filters"] = assay_filter
    final_payload["distinct"] = True
    final_payload["adjust_geno_bins"] = False

    field_names = []
    for f in fields:
        field_names.append(list(f.keys())[0])

    return final_payload, field_names



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import print_function, unicode_literals, division, absolute_import

import unittest, time, json, re, os
import dateutil.parser
import dxpy
from dxpy import AppError, AppInternalError, DXError, DXFile, DXRecord
from dxpy.utils import (exec_utils, genomic_utils, response_iterator, get_futures_threadpool, DXJSONEncoder,
                        normalize_timedelta, normalize_time_input, config, Nonce)
from dxpy.utils.exec_utils import DXExecDependencyInstaller
from dxpy.utils.pretty_print import flatten_json_array
import dxpy_testutil as testutil
from dxpy.system_requirements import SystemRequirementsDict

# TODO: unit tests for dxpy.utils.get_field_from_jbor, get_job_from_jbor, is_job_ref


class TestErrorSanitizing(unittest.TestCase):
    def test_error_sanitizing(self):
        # ASCII str
        self.assertEqual(exec_utils._safe_unicode(ValueError("foo")), "foo")
        # UTF-8 encoded str
        self.assertEqual(exec_utils._safe_unicode(ValueError("crème".encode("utf-8"))),
                         "b'cr\\xc3\\xa8me'")
        # Unicode obj
        self.assertEqual(exec_utils._safe_unicode(ValueError("brûlée")), "br\xfbl\xe9e")
        # Not UTF-8
        expected = "b'Invalid read name: D\\xd1\\xc1n\\xe8x\\xfbs'"
        self.assertEqual(exec_utils._safe_unicode(ValueError("Invalid read name: DÑÁnèxûs".encode("ISO-8859-1"))), expected)

    def test_formatting_exceptions(self):
        self.assertEqual(exec_utils._format_exception_message(ValueError("foo")), "ValueError: foo")
        self.assertEqual(exec_utils._format_exception_message(AppError("foo")), "foo")

class TestGenomicUtils(unittest.TestCase):
    def test_reverse_complement(self):
        self.assertEqual(b"TTTTAAACCG", genomic_utils.reverse_complement(b"CGGTTTAAAA"))
        self.assertEqual(b"TTTTAAACCG", genomic_utils.reverse_complement("CGGTTTAAAA"))
        self.assertEqual(b"TTTTAAACCG", genomic_utils.reverse_complement(b"cggtttaaaa"))
        self.assertEqual(b"TTTTAAACCG", genomic_utils.reverse_complement("cggtttaaaa"))
        self.assertEqual(b"NNNNNTTTTAAACCG", genomic_utils.reverse_complement(b"CGGTTTAAAANNNNN"))
        self.assertEqual(b"NNNNNTTTTAAACCG", genomic_utils.reverse_complement("CGGTTTAAAANNNNN"))
        with self.assertRaises(ValueError):
            genomic_utils.reverse_complement("oops")

class TestResponseIterator(unittest.TestCase):
    def test_basic_iteration(self):
        def task(i, sleep_for=1):
            print("Task", i, "sleeping for", sleep_for)
            time.sleep(sleep_for)
            return i

        def tasks():
            for i in range(8):
                yield task, [i], {"sleep_for": i/8.0}

        for i, res in enumerate(response_iterator(tasks(), get_futures_threadpool(3))):
            self.assertEqual(i, res)
        for i, res in enumerate(response_iterator(tasks(), get_futures_threadpool(5))):
            self.assertEqual(i, res)
        for i, res in enumerate(response_iterator(tasks(), get_futures_threadpool(5), max_active_tasks=2)):
            self.assertEqual(i, res)
        for i, res in enumerate(response_iterator(tasks(), get_futures_threadpool(5), max_active_tasks=6)):
            self.assertEqual(i, res)

    def test_first_chunk_sequentially(self):
        pending_tasks = set()
        complete_tasks = set()

        def task(i):
            pending_tasks.add(i)
            if i >= 1:
                self.assertNotIn(0, pending_tasks, "Task 0 was pending during task %d" % i)
            time.sleep(1)
            complete_tasks.add(i)
            pending_tasks.remove(i)

        def tasks():
            for i in range(8):
                yield task, [i], {}

        #
        # |---0---|
        #           |---1---|
        #           |---2---|
        #           |---3---|
        #
        # Assert that 0 doesn't overlap with 1, 2, or 3
        list(response_iterator(tasks(), get_futures_threadpool(8), do_first_task_sequentially=True))


class TestDXUtils(unittest.TestCase):
    def test_dxjsonencoder(self):
        f = DXFile("file-" + "x"*24, project="project-" + "y"*24)
        r = DXRecord("record-" + "r"*24, project="project-" + "y"*24)
        data = {"a": [{"b": f}, r]}
        serialized = json.dumps(data, cls=DXJSONEncoder)
        self.assertEqual(serialized,
                         '{"a": [{"b": {"$dnanexus_link": "file-xxxxxxxxxxxxxxxxxxxxxxxx"}}, {"$dnanexus_link": "record-rrrrrrrrrrrrrrrrrrrrrrrr"}]}')

class TestEDI(DXExecDependencyInstaller):
    __test__ = False
    def __init__(self, *args, **kwargs):
        self.command_log, self.message_log = [], []
        DXExecDependencyInstaller.__init__(self, *args, **kwargs)

    def run(self, cmd, **kwargs):
        self.command_log.append(cmd)

    def log(self, message, **kwargs):
        self.message_log.append(message)

class TestDXExecDependsUtils(testutil.DXTestCaseCompat):
    def get_edi(self, run_spec, job_desc=None):
        return TestEDI(executable_desc={"runSpec": run_spec}, job_desc=job_desc if job_desc else {})

    def test_install_bundled_dependencies(self):
        bundled_depends_by_region = {
            "aws:us-east-1": [
                {"name": "asset.east", "id": {"$dnanexus_link": "file-asseteast"}}
            ],

            "azure:westus": [
                {"name": "asset.west", "id": {"$dnanexus_link": "file-assetwest"}}
            ]
        }
        edi = self.get_edi({"bundledDependsByRegion": bundled_depends_by_region},
                           job_desc={"region": "aws:us-east-1"})
        with self.assertRaisesRegex(DXError, 'file-asseteast'):
            # Asserts that we attempted to download the correct file.
            edi.install()
        edi = self.get_edi({"bundledDependsByRegion": bundled_depends_by_region},
                           job_desc={"region": "azure:westus"})
        with self.assertRaisesRegex(DXError, 'file-assetwest'):
            edi.install()
        with self.assertRaisesRegex(KeyError, 'aws:eu-central-1'):
            self.get_edi({"bundledDependsByRegion": bundled_depends_by_region},
                         job_desc={"region": "aws:eu-central-1"})

    def test_dx_execdepends_installer(self):
        def assert_cmd_ran(edi, regexp):
            self.assertRegex("\n".join(edi.command_log), regexp)

        def assert_log_contains(edi, regexp):
            self.assertRegex("\n".join(edi.message_log), regexp)

        with self.assertRaisesRegex(AppInternalError, 'Expected field "runSpec" to be present'):
            DXExecDependencyInstaller({}, {})

        with self.assertRaisesRegex(AppInternalError, 'Expected field "name" to be present'):
            self.get_edi({"dependencies": [{"foo": "bar"}]})

        edi = self.get_edi({"dependencies": [{"name": "foo", "package_manager": "cran", "version": "1.2.3"}]})
        edi.install()
        assert_cmd_ran(edi, "R -e .+ install.packages.+devtools.+install_version.+foo.+version.+1.2.3")

        with self.assertRaisesRegex(AppInternalError, 'does not have a "url" field'):
            self.get_edi({"dependencies": [{"name": "foo", "package_manager": "git"}]})

        edi = self.get_edi({"execDepends": [{"name": "git"}], "dependencies": [{"name": "tmux"}]})
        edi.install()
        assert_cmd_ran(edi, "apt-get install --yes --no-install-recommends git tmux")

        edi = self.get_edi({"execDepends": [], "bundledDepends": [], "dependencies": []})
        edi.install()

        edi = self.get_edi({"dependencies": [{"name": "pytz", "package_manager": "pip", "version": "2014.7"},
                                             {"name": "certifi", "package_manager": "pip", "stages": ["main"]},
                                             {"name": "tmux", "package_manager": "apt"},
                                             {"name": "rake", "package_manager": "gem", "version": "10.3.2"},
                                             {"name": "nokogiri", "package_manager": "gem", "stages": ["main"]},
                                             {"name": "Module::Provision", "package_manager": "cpan", "version": "0.36.1"},
                                             {"name": "LWP::MediaTypes", "package_manager": "cpan"},
                                             {"name": "RJSONIO", "package_manager": "cran"},
                                             {"name": "plyr", "package_manager": "cran", "version": "1.8.1"},
                                             {"name": "ggplot2", "package_manager": "cran", "stages": ["main"],
                                              "version": "1.0.1"},
                                             {"name": "r1", "id": {"$dnanexus_link": "record-123"}},
                                             {"name": "g1",
                                              "package_manager": "git",
                                              "url": "https://github.com/dnanexus/oauth2-demo"},
                                             {"name": "g2",
                                              "package_manager": "git",
                                              "url": "https://github.com/dnanexus/bwa",
                                              "tag": "production",
                                              "destdir": "/tmp/ee-edi-test-bwa",
                                              "buld_commands": "echo build bwa here",
                                              "stages": ["main"]}]})
        edi.install()
        assert_cmd_ran(edi, re.escape("pip install --upgrade pytz==2014.7 certifi"))
        assert_cmd_ran(edi, "apt-get install --yes --no-install-recommends tmux")
        assert_cmd_ran(edi, re.escape("gem install rake --version 10.3.2 && gem install nokogiri"))
        assert_cmd_ran(edi, "R -e .+ install.packages.+\"RJSONIO\".+install_version.+\"ggplot2\".+version=\"1.0.1\"")
        assert_log_contains(edi, 'Skipping bundled dependency "r1" because it does not refer to a file')
        assert_cmd_ran(edi, re.escape("cd $(mktemp -d) && git clone https://github.com/dnanexus/oauth2-demo"))
        assert_cmd_ran(edi, "cd /tmp/ee-edi-test-bwa && git clone https://github.com/dnanexus/bwa")
        assert_cmd_ran(edi, "git checkout production")

        edi = self.get_edi({"execDepends": [{"name": "w00t", "stages": ["foo", "bar"]},
                                            {"name": "f1", "id": {"$dnanexus_link": "file-123"}, "stages": ["xyzzt"]}]})
        edi.install()
        self.assertNotRegex("\n".join(edi.command_log), "w00t")
        for name in "w00t", "f1":
            assert_log_contains(edi,
                                r"Skipping dependency {} because it is inactive in stage \(function\) main".format(name))

        edi = self.get_edi({"execDepends": [{"name": "git", "stages": ["foo", "bar"]}]},
                           job_desc={"function": "foo"})
        edi.install()
        assert_cmd_ran(edi, "apt-get install --yes --no-install-recommends git")

        # Job describe dict must contain "region" if the run specification
        # contains "bundledDependsByRegion".
        bundled_depends_by_region = {
            "aws:us-east-1": [
                {"name": "asset.east", "id": {"$dnanexus_link": "file-asseteast"}}
            ]
        }
        with self.assertRaisesRegex(DXError, 'region.*job description'):
            self.get_edi({"bundledDependsByRegion": bundled_depends_by_region})

        bundled_depends_by_region = {
            "aws:us-east-1": []
        }
        with self.assertRaisesRegex(DXError, 'region.*job description'):
            self.get_edi({"bundledDependsByRegion": bundled_depends_by_region})

        # Job describe dict may specify or omit "region" if
        # "bundledDependsByRegion" is not in run specification.
        edi = self.get_edi({}, job_desc={"region": "azure:westus"})
        edi.install()

        edi = self.get_edi({}, job_desc={})
        edi.install()


class TestTimeUtils(unittest.TestCase):
    def test_normalize_timedelta(self):
        for i, o in (("-15", -15000),
                     ("15", 15000),
                     ("15s", 15000),
                     ("0", 0),
                     ("0w", 0),
                     ("1m", 1000*60),
                     ("1M", 1000*60*60*24*30),
                     ("-1w", -1000*60*60*24*7)):
            self.assertEqual(normalize_timedelta(i), o)

    def test_strings_vs_ints(self):
        # This is potentially confusing; it would be nice to have two
        # separate methods, or a flag to control the interpretation of
        # the field when a string with no suffix is supplied.

        # "dx login" can supply this form
        self.assertEqual(normalize_time_input("1414141414", default_unit='s'), 1414141414000)   # interpreted as sec
        # "dx find ... --created-*" can supply this form
        self.assertEqual(normalize_time_input("1234567890"), 1234567890)  # interpreted as ms
        self.assertEqual(normalize_time_input(1414141414000), 1414141414000)  # interpreted as ms

    def test_normalize_time_input(self):
        # TODO: Add tests for negative time inputs e.g. "-12345", -12345, "-5d"

        for i, o in ((12345678, 12345678),
                     ("0", 0),
                     ("12345678", 12345678),
                     ("15s", 15 * 1000),
                     ("1d", (24 * 60 * 60 * 1000)),
                     ("0w", 0),
                     ("2015-10-01", int(time.mktime(dateutil.parser.parse("2015-10-01").timetuple()) * 1000))):
            self.assertEqual(normalize_time_input(i), o)

        # Test default_unit='s'
        for i, o in ((12345678, 12345678 * 1000),
                     ("12345678", 12345678 * 1000),
                     ("15s", 15 * 1000),
                     ("1d", (24 * 60 * 60 * 1000)),
                     ("0w", 0),
                     ("2015-10-01", int(time.mktime(dateutil.parser.parse("2015-10-01").timetuple()) * 1000))):
            self.assertEqual(normalize_time_input(i, default_unit='s'), o)

        with self.assertRaises(ValueError):
            normalize_time_input("1223*")
        with self.assertRaises(ValueError):
            normalize_time_input("12345", default_unit='h')
        with self.assertRaises(ValueError):
            normalize_time_input(12345, default_unit='h')
        with self.assertRaises(ValueError):
            normalize_time_input("1234.5678")
        with self.assertRaises(ValueError):
            normalize_time_input(1234.5678)


class TestDXConfig(unittest.TestCase):
    def test_dxconfig(self):
        environ_backup = os.environ.copy()
        try:
            c = config.DXConfig()
            for var in c.VAR_NAMES:
                value = '{"foo": "bar"}'
                c[var] = value
                if var in c.CORE_VAR_NAMES - {"DX_SECURITY_CONTEXT", "DX_WORKSPACE_ID"}:
                    self.assertEqual(getattr(dxpy, var.lstrip("DX_")), value)
                elif var == "DX_SECURITY_CONTEXT":
                    self.assertEqual(json.dumps(dxpy.SECURITY_CONTEXT), value)
                c.write(var, None)
                del c[var]
            c.update(DX_CLI_WD="/wd")
            self.assertIn("DX_CLI_WD", c)
            self.assertEqual(c["DX_CLI_WD"], "/wd")
            self.assertEqual(c.pop("DX_CLI_WD"), "/wd")
            self.assertEqual(c.pop("DX_CLI_WD", None), None)
            self.assertIn("DXConfig object at", repr(c))
            c.update(c.defaults)
            self.assertEqual(len(c), len(list(c)))
            del c[list(c.defaults.keys())[0]]
            self.assertEqual(len(c), len(list(c)))
            dxpy.config["DX_PROJECT_CONTEXT_NAME"] = None
            self.assertEqual(os.environ["DX_PROJECT_CONTEXT_NAME"], "")
        finally:
            os.environ.update(environ_backup)
            dxpy.config.__init__(suppress_warning=True)

class TestPrettyPrint(unittest.TestCase):
    def test_flatten_json_array(self):
        json_string = (
            '{\n'
            '  "arr": [\n'
            '    "one",\n'
            '    2,\n'
            '    3.4,\n'
            '    "\\"five\\""\n'
            '  ],\n'
            '  "foo": {\n'
            '    "arr": [\n'
            '      "six",\n'
            '      7.8,\n'
            '      9,\n'
            '      "t\\"en"\n'
            '    ]\n'
            '  }\n'
            '}'
        )
        flattened_json_string_ref = (
            '{\n'
            '  "arr": ["one", 2, 3.4, "\\"five\\""],\n'
            '  "foo": {\n'
            '    "arr": ["six", 7.8, 9, "t\\"en"]\n'
            '  }\n'
            '}'
        )
        flatten_json_array(json_string, "arr")
        self.assertEqual(flattened_json_string_ref, flatten_json_array(json_string, "arr"))


class TestNonceGeneration(unittest.TestCase):
    def test_nonce_generator(self):
        nonce_list = []
        for i in range(0, 100):
            nonce_list.append(str(Nonce()))

        for nonce in nonce_list:
            self.assertTrue(len(nonce) > 0)
            self.assertTrue(len(nonce) <= 128)
            self.assertEqual(nonce_list.count(nonce), 1)

    def test_input_updater(self):
        input_params = {"p1": "v1", "p2": "v2"}
        updated_input = Nonce.update_nonce(input_params)
        self.assertIn("nonce", updated_input)

        nonce = str(Nonce())
        input_params.update({"nonce": nonce})
        updated_input = Nonce.update_nonce(input_params)
        self.assertIn("nonce", updated_input)
        self.assertEqual(nonce, updated_input["nonce"])

class TestSystemRequirementsDict(unittest.TestCase):

    def test_add(self):
        d1 = {'a': {'x': 'pqr'}}
        d2 = {'a': {'y': 'lmn'}, 'b': {'y': 'rst'}}
        srd1 = SystemRequirementsDict(d1)
        srd2 = SystemRequirementsDict(d2)
        added = srd1 + srd2
        expected = SystemRequirementsDict({'a': {'x': 'pqr', 'y': 'lmn'}, 'b': {'y': 'rst'}})
        self.assertDictEqual(expected.entrypoints, added.entrypoints)

        d1 = {}
        d2 = {'a': {'y': 'lmn'}, 'b': {'y': 'rst'}}
        srd1 = SystemRequirementsDict(d1)
        srd2 = SystemRequirementsDict(d2)
        added = srd1 + srd2
        expected = SystemRequirementsDict({'a': {'y': 'lmn'}, 'b': {'y': 'rst'}})
        self.assertDictEqual(expected.entrypoints, added.entrypoints)

        d1 = None
        d2 = {'a': {'y': 'lmn'}, 'b': {'y': 'rst'}}
        srd1 = SystemRequirementsDict(d1)
        srd2 = SystemRequirementsDict(d2)
        added = srd1 + srd2
        expected = SystemRequirementsDict({'a': {'y': 'lmn'}, 'b': {'y': 'rst'}})
        self.assertDictEqual(expected.entrypoints, added.entrypoints)

        d1 = None
        d2 = None
        srd1 = SystemRequirementsDict(d1)
        srd2 = SystemRequirementsDict(d2)
        added = srd1 + srd2
        expected = SystemRequirementsDict(None)
        self.assertEqual(expected.entrypoints, added.entrypoints)

    def test_override_cluster_spec_for_app_with_named_entrypoint(self):
        bootstrap_code = "import sys\n"
        cluster_spec_with_bootstrap = {"type": "spark",
                                            "version": "2.4.0",
                                            "initialInstanceCount": 2,
                                            "bootstrapScript": bootstrap_code}
        cluster_spec_no_bootstrap = {"type": "spark",
                                     "version": "2.4.0",
                                     "initialInstanceCount": 3}
        app_sys_reqs = {"main": {
                         "instanceType": "mem2_hdd2_x1",
                         "clusterSpec": cluster_spec_with_bootstrap},
                        "cluster_2": {
                         "instanceType": "mem2_hdd2_x4",
                         "clusterSpec": cluster_spec_no_bootstrap}
                       }
        app_srd = SystemRequirementsDict.from_sys_requirements(app_sys_reqs, _type='clusterSpec')

        # pass instance count with specific entry point
        runtime_srd = SystemRequirementsDict.from_instance_count({"cluster_2": "4"})
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        self.assertEqual(cluster_spec_srd.entrypoints['cluster_2']["clusterSpec"]["initialInstanceCount"], 4)
        self.assertEqual(cluster_spec_srd.entrypoints['cluster_2']["clusterSpec"]["version"], "2.4.0")
        self.assertEqual(cluster_spec_srd.entrypoints['cluster_2'].get('instanceType'), None)

        # pass instance count for all entry points ("*")
        runtime_srd = SystemRequirementsDict.from_instance_count(5)
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        self.assertEqual(cluster_spec_srd.entrypoints['cluster_2']["clusterSpec"]["initialInstanceCount"], 5)
        self.assertEqual(cluster_spec_srd.entrypoints['main']["clusterSpec"]["initialInstanceCount"], 5)
        self.assertEqual(cluster_spec_srd.entrypoints['main']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(cluster_spec_srd.entrypoints['main'].get('instanceType'), None)
        self.assertEqual(cluster_spec_srd.entrypoints.get('*'), None)

        # pass instance count together with instance type
        runtime_srd = SystemRequirementsDict.from_instance_count({"main": "6"})
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        instance_type_srd = SystemRequirementsDict.from_instance_type({"main": "mem1_ssd1_x2"})
        added = (cluster_spec_srd + instance_type_srd).as_dict()
        self.assertEqual(added['main']["clusterSpec"]["initialInstanceCount"], 6)
        self.assertEqual(added['main']["clusterSpec"]["version"], "2.4.0")
        self.assertEqual(added['main']["instanceType"], "mem1_ssd1_x2")
        self.assertTrue("*" not in added)
        self.assertTrue("cluster_2" not in added)

    def test_override_cluster_spec_for_app_with_wildcard_entrypoint(self):
        bootstrap_code = "import sys\n"
        app_sys_reqs = {"*": {
                            "instanceType": "mem2_hdd2_x1",
                            "clusterSpec": {"type": "spark",
                                            "version": "2.4.0",
                                            "initialInstanceCount": 22,
                                            "bootstrapScript": bootstrap_code}},
                        "other": {
                            "instanceType": "mem2_hdd2_x2",
                            "clusterSpec": {"type": "spark",
                                            "version": "2.4.0",
                                            "initialInstanceCount": 33,
                                            "bootstrapScript": bootstrap_code}
                        }}
        app_srd = SystemRequirementsDict.from_sys_requirements(app_sys_reqs, _type='clusterSpec')

        # pass instance count with "*" entry point
        runtime_srd = SystemRequirementsDict.from_instance_count(8)
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        self.assertEqual(cluster_spec_srd.entrypoints['*']["clusterSpec"]["initialInstanceCount"], 8)
        self.assertEqual(cluster_spec_srd.entrypoints['*']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(cluster_spec_srd.entrypoints['other']["clusterSpec"]["initialInstanceCount"], 8)
        self.assertEqual(cluster_spec_srd.entrypoints['other']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(cluster_spec_srd.entrypoints['*'].get('instanceType'), None)

        # pass instance count with a named entry point
        runtime_srd = SystemRequirementsDict.from_instance_count({"main": 77})
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        self.assertEqual(cluster_spec_srd.entrypoints['main']["clusterSpec"]["initialInstanceCount"], 77)
        self.assertEqual(cluster_spec_srd.entrypoints['main']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(cluster_spec_srd.entrypoints['main'].get('instanceType'), None)

        # pass instance count with a named entry point and instance type
        runtime_srd = SystemRequirementsDict.from_instance_count({"main": "42"})
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        instance_type_srd = SystemRequirementsDict.from_instance_type({"main": "mem1_ssd1_x2"})
        added = (cluster_spec_srd + instance_type_srd).as_dict()
        self.assertEqual(added['main']["clusterSpec"]["initialInstanceCount"], 42)
        self.assertEqual(added['main']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(added['main']["instanceType"], "mem1_ssd1_x2")

        # pass instance count with a named and wildcard entry point and instance type
        runtime_srd = SystemRequirementsDict.from_instance_count({"main": "42", "*": "52"})
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        instance_type_srd = SystemRequirementsDict.from_instance_type({"main": "mem1_ssd1_x2"})
        added = (cluster_spec_srd + instance_type_srd).as_dict()
        self.assertEqual(added['main']["clusterSpec"]["initialInstanceCount"], 42)
        self.assertEqual(added['main']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(added['main']["instanceType"], "mem1_ssd1_x2")
        self.assertEqual(added['*']["clusterSpec"]["initialInstanceCount"], 52)
        self.assertTrue("instanceType" not in added['*'])

    def test_override_cluster_spec_for_app_with_non_spark_entrypoint(self):
        bootstrap_code = "import sys\n"
        app_sys_reqs = {"non_spark_entry": {
                            "instanceType": "mem2_hdd2_x1"},
                        "spark_entry": {
                            "clusterSpec": {"type": "spark",
                                            "version": "2.4.0",
                                            "initialInstanceCount": 2,
                                            "bootstrapScript": bootstrap_code}}}
        app_srd = SystemRequirementsDict.from_sys_requirements(app_sys_reqs, _type='clusterSpec')

        # pass instance count with "*" entry point
        runtime_srd = SystemRequirementsDict.from_instance_count(8)
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        self.assertEqual(cluster_spec_srd.entrypoints['spark_entry']["clusterSpec"]["initialInstanceCount"], 8)
        self.assertEqual(cluster_spec_srd.entrypoints['spark_entry']["clusterSpec"]["bootstrapScript"], bootstrap_code)
        self.assertEqual(cluster_spec_srd.entrypoints['spark_entry'].get('instanceType'), None)
        self.assertTrue("*" not in cluster_spec_srd.entrypoints)
        self.assertTrue("non_spark_entry" not in cluster_spec_srd.entrypoints)

        # pass instance count and instance type
        runtime_srd = SystemRequirementsDict.from_instance_count({"spark_entry": "422", "*": "522"})
        cluster_spec_srd = app_srd.override_cluster_spec(runtime_srd)
        instance_type_srd = SystemRequirementsDict.from_instance_type({"*": "mem1_ssd1_x2",
                                                                  "spark_entry": "mem1_ssd1_x4",
                                                                  "no_spark_entry": "mem1_ssd1_x8"})
        added = (cluster_spec_srd + instance_type_srd).as_dict()
        expected = {
                    'spark_entry': {
                            'clusterSpec': {
                                    'bootstrapScript': 'import sys\n',
                                    'version': '2.4.0',
                                    'type': 'spark',
                                    'initialInstanceCount': 422},
                            'instanceType': 'mem1_ssd1_x4'},
                    '*': {
                            'instanceType': 'mem1_ssd1_x2'},
                    'no_spark_entry': {
                            'instanceType': 'mem1_ssd1_x8'
                    }}
        self.assertEqual(added, expected)

    def test_from_instance_type(self):
        d1 = None
        d2 = None
        srd1 = SystemRequirementsDict(d1)
        srd2 = SystemRequirementsDict.from_instance_type(d2)
        added = srd1 + srd2
        expected = SystemRequirementsDict(None)
        self.assertEqual(expected.entrypoints, added.entrypoints)

if __name__ == '__main__':
    unittest.main()
